from babelfont import load


def test_load_open_shape():
    font = load("tests/data/GlyphsFileFormatv3.glyphs")
    assert font.glyphs["A"].layers[1].shapes[0].closed == False


def test_designspace():
    font = load("tests/data/Designspace.glyphs")
    assert len(font.axes) == 1
    assert font.axes[0].name.get_default() == "Weight"
    assert font.axes[0].tag == "wght"
    # Axes values are in userspace units
    assert font.axes[0].minimum == 100
    assert font.axes[0].maximum == 600
    # Master locations are in designspace units
    assert font.masters[0].location == {"wght": 1}
    assert font.masters[1].location == {"wght": 199}
    # Instance locations are in designspace units
    assert font.instances[0].location == {"wght": 1}
    assert font.instances[1].location == {"wght": 7}



import logging
from importlib.resources import files
from xml.etree import ElementTree

from babelfont import Font

logger = logging.getLogger(__name__)


def bake_in_glyphdata(font: Font, args=None):
    logger.info("Baking in glyph data")

    if not args or "file" not in args:
        data_dir = files("babelfont.data")
        file = data_dir / "GlyphData.xml"
    else:
        file = args["file"]

    parsed = ElementTree.parse(open(file, "rb")).getroot()
    for glyph in parsed:
        data = glyph.attrib
        if data["name"] not in font.glyphs:
            continue
        myglyph = font.glyphs[data["name"]]
        if "production" in data and not myglyph.production_name:
            myglyph.production_name = data["production"]
        if "category" in data and not myglyph.category:
            if data["category"] == "Mark":
                myglyph.category = "mark"



import logging
from babelfont.Font import Font

logger = logging.getLogger(__name__)


def zero_mark_widths(font: Font, args=None):
    logger.info("Zeroing mark widths")
    for glyph in font.glyphs:
        if glyph.category != "mark":
            continue
        for layer in glyph.layers:
            layer.width = 0



import logging
from typing import List

from babelfont.Font import Font
from babelfont.Glyph import GlyphList

logger = logging.getLogger(__name__)


def _replaceFeature(font: Font, value: str):
    tag, code = value.split(";", 1)
    logger.info(f"Replacing feature '{tag}'")
    font.features.features = [
        (key, code if key == tag else value) for key, value in font.features.features
    ]


def _removeFeatures(font: Font, value: List[str]):
    logger.info(f"Removing features '{', '.join(value)}'")
    font.features.features = [
        (key, value) for key, value in font.features.features if key not in value
    ]


def _removeGlyphs(font: Font, value: List[str]):
    logger.info(f"Removing glyphs")
    # for glyph in value:
    #     font.glyphs.pop(glyph)


def _decomposeGlyphs(font: Font, value: List[str]):
    logger.info(f"Decomposing glyphs")
    for glyph in font.glyphs:
        if glyph.name in value:
            # logger.info(f"Decomposing " + glyph.name)
            for layer in glyph.layers:
                layer.decompose()


def _renameGlyphs(font: Font, value: List[str]):
    logger.info(f"Renaming glyphs")
    for mapping in value:
        oldname, newname = mapping.split("=", 1)
        if newname in font.glyphs and oldname in font.glyphs:
            # Swap any unicode mappings
            (font.glyphs[oldname].codepoints, font.glyphs[newname].codepoints) = (
                font.glyphs[newname].codepoints,
                font.glyphs[oldname].codepoints,
            )
            font.glyphs[newname].name = oldname
            font.glyphs[oldname].name = newname
            font.glyphs = GlyphList({glyph.name: glyph for glyph in font.glyphs})


CUSTOM_PARAMETER_APPLIER = {
    "Replace Feature": _replaceFeature,
    "Remove Features": _removeFeatures,
    "Remove Glyphs": _removeGlyphs,
    "Decompose Glyphs": _decomposeGlyphs,
    "Rename Glyphs": _renameGlyphs,
}


def apply_custom_parameters(font: Font, args=None):
    if "instance" in args:
        root = None
        for instance in font.instances:
            short_instance_name = instance.name.get_default()
            long_instance_name = (
                instance.familyName.get_default() + " " + short_instance_name
            )
            if args["instance"] in (short_instance_name, long_instance_name):
                root = instance._formatspecific.get("com.glyphsapp", {})
                break
        if not root:
            raise ValueError(f"Instance '{args['instance']}' not found")
    else:
        root = font._formatspecific.get("com.glyphsapp", {})

    for cp in root.get("customParameters", []):
        name = cp.get("name")
        value = cp.get("value")
        if name in CUSTOM_PARAMETER_APPLIER:
            CUSTOM_PARAMETER_APPLIER[name](font, value)



from babelfont import load
from babelfont.fontFilters.rename import rename_glyphs


def test_rename():
    font = load("tests/data/NotoSerifMakasar-Regular.ufo")
    assert "uni11EF0" in font.features.to_fea()
    rename_glyphs(
        font, {"mapping": {"uni11EF0": "sa-makasar", "uni11EF3": "iVowel-makasar"}}
    )
    assert "sa-makasar" in font.features.to_fea()
    assert (
        "markClass iVowel-makasar <anchor -280 622> @_MARKCLASS_0_lookup_0;"
        in font.features.to_fea()
    )
    assert "sa-makasar" in font.glyphs
    assert "uni11EF0" not in font.glyphs


def test_rename_production():
    font = load("tests/data/GlyphsFileFormatv3.glyphs")
    rename_glyphs(font, {"production": True})
    assert "Smily" not in font.glyphs
    assert "someSmily" in font.glyphs


def test_rename_nested():
    font = load("tests/data/NotoSerifMakasar-Regular.ufo")
    font.features.features = [("ccmp", "lookup foo { sub uni11EF0 by uni11EF3; } foo;")]
    rename_glyphs(
        font, {"mapping": {"uni11EF0": "sa-makasar", "uni11EF3": "iVowel-makasar"}}
    )
    assert "sub sa-makasar by iVowel-makasar;" in font.features.to_fea()


def test_rename_contextual():
    font = load("tests/data/NotoSerifMakasar-Regular.ufo")

    font.features.features = [
        (
            "kern",
            """
    pos uni11EF0 [uni11EF3]' 50;
    """,
        )
    ]
    rename_glyphs(
        font, {"mapping": {"uni11EF0": "sa-makasar", "uni11EF3": "iVowel-makasar"}}
    )
    assert "pos sa-makasar [iVowel-makasar]' 50" in font.features.features[0][1]



from io import StringIO
import re
from dataclasses import dataclass, field
from typing import Any, Dict, List, Tuple

from babelfont import Font
from fontTools.feaLib import ast
from fontTools.feaLib.parser import Parser, SymbolTable

from .BaseObject import BaseObject

PREFIX_MARKER = "# Prefix: "
PREFIX_RE = re.compile(r"# Prefix: (.*)")


@dataclass
class Features(BaseObject):
    """A representation of the OpenType feature code."""

    classes: Dict[str, List[str]] = field(
        default_factory=dict,
        metadata={
            "separate_items": True,
            "description": "A dictionary of classes. Each group is a list of glyph names or class names. The key should not start with @.",
        },
    )
    prefixes: Dict[str, str] = field(
        default_factory=dict,
        metadata={
            "separate_items": True,
            "description": "A dictionary of OpenType lookups and other feature code to be placed before features are defined. The keys are user-defined names, the values are AFDKO feature code.",
        },
    )
    features: List[Tuple[str, str]] = field(
        default_factory=list,
        metadata={
            "separate_items": True,
            "description": "A list of OpenType feature code, expressed as a tuple (feature tag, code).",
        },
    )

    @classmethod
    def from_fea(cls, fea: str, glyphNames=()) -> "Features":
        """Load features from a .fea file."""
        parsed = Parser(
            StringIO(fea), followIncludes=False, glyphNames=glyphNames
        ).parse()
        features = Features()
        currentPrefix = "anonymous"
        for statement in parsed.statements:
            if isinstance(statement, ast.GlyphClassDefinition):
                statement.glyphs.asFea()  # This builds .original
                if statement.glyphs.original:
                    features.classes[statement.name] = [
                        ast.asFea(x) for x in statement.glyphs.original
                    ]
                else:
                    features.classes[statement.name] = list(statement.glyphs.glyphs)
            elif isinstance(statement, ast.FeatureBlock):
                features.features.append((statement.name, ast.Block.asFea(statement)))
            elif isinstance(statement, ast.Comment) and (
                m := re.match(PREFIX_RE, statement.text)
            ):
                currentPrefix = m.group(1)
            else:
                if currentPrefix not in features.prefixes:
                    features.prefixes[currentPrefix] = ""
                features.prefixes[currentPrefix] += statement.asFea()
        return features

    def to_fea(self) -> str:
        """Dump features to a .fea file."""
        fea = ""
        for name, glyphs in self.classes.items():
            fea += f"@{name} = [{' '.join(glyphs)}];\n"
        for prefix, code in self.prefixes.items():
            if prefix != "anonymous":
                fea += f"# Prefix: {prefix}\n"
            fea += code + "\n"
        for name, code in self.features:
            fea += f"feature {name} {{\n{code}\n}} {name};\n"
        return fea

    def as_ast(self, font: Font) -> Dict[str, Any]:
        rv = {
            "prefixes": {},
            "features": [],
        }
        glyphnames = font.glyphs.keys()
        lookups = SymbolTable()
        glyphclasses = SymbolTable()
        for name, glyphs in self.classes.items():
            glyphcls = ast.GlyphClass(glyphs=[ast.GlyphName(g) for g in glyphs])
            glyphclass = ast.GlyphClassDefinition(name, glyphcls)
            glyphclasses.define(name, glyphclass)

        for prefix, code in self.prefixes.items():
            parser = Parser(StringIO(code), followIncludes=False, glyphNames=glyphnames)
            parser.lookups_ = lookups
            parser.glyphclasses_ = glyphclasses
            try:
                file = parser.parse()
            except Exception as e:
                raise ValueError(f"Error parsing feature code: {e}\n\nCode was: {code}")
            rv["prefixes"][prefix] = file

        for name, code in self.features:
            if name is not None and not code.startswith("feature " + name):
                code = f"feature {name} {{\n{code}\n}} {name};"
            parser = Parser(StringIO(code), followIncludes=False, glyphNames=glyphnames)
            parser.lookups_ = lookups
            parser.glyphclasses_ = glyphclasses
            try:
                file = parser.parse()
            except Exception as e:
                raise ValueError(f"Error parsing feature code: {e}\n\nCode was: {code}")
            rv["features"].append((name, file))

        return rv



import os
import sys
import pkgutil
import inspect
import importlib
import logging

from babelfont.Font import Font
from babelfont.fontFilters import parse_filter

logger = logging.getLogger(__name__)


class BaseConvertor:
    filename: str
    scratch: object
    font: Font
    compile_only: bool

    suffix = ".XXX"

    LOAD_FILTERS = []
    COMPILE_FILTERS = []
    SAVE_FILTERS = []

    def __init__(self):
        self.logger = logging.getLogger(self.__class__.__name__)

    @classmethod
    def can_load(cls, other, **kwargs):
        return other.filename.endswith(cls.suffix)

    @classmethod
    def can_save(cls, other, **kwargs):
        return other.filename.endswith(cls.suffix)

    @classmethod
    def load(cls, convertor, compile_only=False, filters=True):
        self = cls()
        self.font = Font()
        # Pass on information to child
        self.filename = convertor.filename
        self.scratch = convertor.scratch
        self.compile_only = compile_only
        loaded = self._load()
        if filters:
            for f in cls.LOAD_FILTERS:
                fltr, filterargs = parse_filter(f)
                fltr(loaded, filterargs)

        return loaded

    @classmethod
    def save(cls, obj, convertor, **kwargs):
        self = cls()
        self.font = obj
        # Pass on information to child
        self.filename = convertor.filename
        self.scratch = convertor.scratch
        return self._save()

    def _load(self):
        raise NotImplementedError

    def _save(self):
        raise NotImplementedError


class Convert:
    convertors = []

    @classmethod
    def _load_convertors(cls):
        if cls.convertors:
            return
        convertorpath = os.path.join(
            os.path.dirname(sys.modules[cls.__module__].__file__)
        )
        # Additional plugin path here?
        loaders = pkgutil.iter_modules([convertorpath])
        for loader, module_name, is_pkg in loaders:
            spec = loader.find_spec(module_name)
            _module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(_module)
            classes = [
                x[1]
                for x in inspect.getmembers(_module, inspect.isclass)
                if issubclass(x[1], BaseConvertor) and x[1] is not BaseConvertor
            ]
            cls.convertors.extend(classes)

    def __init__(self, filename):
        self._load_convertors()
        self.filename = filename
        self.scratch = {}

    def load_convertor(self, **kwargs):
        for c in self.convertors:
            if c.can_load(self, **kwargs):
                return c

    def save_convertor(self, **kwargs):
        for c in self.convertors:
            if c.can_save(self, **kwargs):
                return c

    def load(self, **kwargs):
        c = self.load_convertor(**kwargs)
        if not c:
            logger.error("Could not find a convertor from %s", self.filename)
            raise NotImplementedError
        return c.load(self, **kwargs)

    def save(self, obj, **kwargs):
        c = self.save_convertor(**kwargs)
        if not c:
            logger.error("Could not find a convertor to %s", self.filename)
            raise NotImplementedError
        return c.save(obj, self, **kwargs)



import logging

from fontTools.cu2qu.ufo import glyphs_to_quadratic

from babelfont.Font import Font


logger = logging.getLogger(__name__)


def cubic_to_quadratic(font: Font, args: dict = {}):
    logger.info("Converting cubic curves to quadratic")
    reverse_direction = args.get("reverseDirection", True)
    for glyph in font.glyphs:
        master_layers = [l for l in glyph.layers if l._master]
        if not master_layers:
            continue
        try:
            glyphs_to_quadratic(master_layers, reverse_direction=reverse_direction)
        except Exception as e:
            logger.warning(
                "Problem converting glyph %s to quadratic (probably incompatible) ",
                glyph.name,
                # e,
            )



from babelfont.fontFilters.glyphs3fea import _translate
from babelfont import Font, Axis


def test_translate():
    testfont = Font()
    testfont.axes = [Axis(None, "wght", None, 400, 600, 900)]
    code = "sub a by b;\n#ifdef VARIABLE\ncondition 600 < wght < 900; sub dollar by dollar.alt;\n#endif"
    newcode = _translate(code, {"index": 0, "font": testfont, "tag": "rclt"})
    assert (
        newcode
        == "sub a by b;\n;} rclt;conditionset __condition_rclt_1 {wght 600 900;} __condition_rclt_1;variation rclt __condition_rclt_1 { sub dollar by dollar.alt;\n"
    )

    code = "sub a by b;\n#ifdef VARIABLE\ncondition wght < 900; sub dollar by dollar.alt;\n#endif"
    newcode = _translate(code, {"index": 0, "font": testfont, "tag": "rclt"})
    assert (
        newcode
        == "sub a by b;\n;} rclt;conditionset __condition_rclt_1 {wght 400 900;} __condition_rclt_1;variation rclt __condition_rclt_1 { sub dollar by dollar.alt;\n"
    )



import logging
import uuid
from collections import defaultdict
from typing import Dict, Optional

from babelfont import Font, Layer, Master

logger = logging.getLogger(__name__)


def promote_intermediate_layers(font: Font, _args: dict):
    # Intermediate layers are sparse masters. To handle intermediate
    # layers, we're going to find all the intermediate layers situated
    # at the same point in the designspace, and create a sparse master
    # at that point.
    additional_masters = defaultdict(list)

    for glyph in font.glyphs:
        newlayers = []
        for layer in glyph.layers:
            if loc := intermediate_location(layer, font):
                additional_masters[tuple(loc.items())].append(layer)
            else:
                newlayers.append(layer)
        glyph.layers = newlayers

    if not additional_masters:
        return
    logger.info("Promoting intermediate layers to sparse masters")

    # Create the masters and associate the layers with them
    for loc_tuple, layers in additional_masters.items():
        loc = dict(loc_tuple)
        if any(master.location == loc for master in font.masters):
            _master_exists_error(loc, layers)
            continue
        master = Master(
            name=str(loc), id=str(uuid.uuid1()), location=loc, sparse=True, font=font
        )
        font.masters.append(master)
        for layer in layers:
            layer._master = master.id
            layer.id = master.id
            # Remove the intermediate coordinate, making it a regular layer
            del layer._formatspecific["com.glyphsapp"]["attr"]["coordinates"]


# We know a layer is "intermediate" if it has a Glyphs-specific attribute
# ["attr"]["coordinates"]. The coordinates are the location of the sparse master;
# just the raw values, so we zip them with the axis tags in order.
def intermediate_location(l: Layer, f: Font) -> Optional[Dict[str, float]]:
    if not l._formatspecific or not l._formatspecific["com.glyphsapp"]:
        return
    attr = l._formatspecific["com.glyphsapp"].get("attr")
    if attr and "coordinates" in attr:
        return {axis.tag: coord for axis, coord in zip(f.axes, attr["coordinates"])}


def _master_exists_error(loc, layers):
    glyphs = ", ".join(sorted(layer.glyph.name for layer in layers))
    pl = "s" if len(layers) > 1 else ""
    logger.error(
        f"Glyph{pl} {glyphs} had an 'intermediate' layer at {loc}, but a master already exists"
    )



from typing import Tuple

from .background import decompose_backgrounds, zero_background_width
from .cu2qu import cubic_to_quadratic
from .customParameters import apply_custom_parameters
from .decomposeMixed import decompose_mixed_glyphs
from .dropUnexported import drop_unexported_glyphs
from .fillOpentype import fill_opentype_values
from .glyphDataXML import bake_in_glyphdata
from .glyphs3fea import translate_glyphs3_fea
from .intermediateLayer import promote_intermediate_layers
from .marks import zero_mark_widths
from .rename import rename_glyphs

FILTERS = {
    "zeroBackgroundWidths": zero_background_width,
    "decomposeBackgrounds": decompose_backgrounds,
    "zeroMarkWidths": zero_mark_widths,
    "renameGlyphs": rename_glyphs,
    "decomposeMixedGlyphs": decompose_mixed_glyphs,
    "dropUnexportedGlyphs": drop_unexported_glyphs,
    "cubicToQuadratic": cubic_to_quadratic,
    "glyphData": bake_in_glyphdata,
    "fillOpentypeValues": fill_opentype_values,
    "intermediateLayers": promote_intermediate_layers,
    "glyphsVariableFeatures": translate_glyphs3_fea,
    "applyCustomParameters": apply_custom_parameters,
}


def parse_filter(input: str) -> Tuple[str, dict]:
    if ":" in input:
        filtername, args = input.split(":", 1)
        args = dict([arg.split("=") for arg in args.split(",")])
    else:
        filtername = input
        args = {}
    if filtername not in FILTERS:
        raise ValueError(f"Unknown filter {filtername}")
    return FILTERS[filtername], args



from collections import defaultdict
import re
import json
import uuid

from fontTools.misc.transform import Transform

from babelfont import Anchor, Axis, Glyph, Layer, Master, Node, Shape, Guide
from babelfont.Glyph import GlyphList
from babelfont.BaseObject import Color, I18NDictionary
from babelfont.convertors import BaseConvertor

tags = {
    "Weight": "wght",
    "Width": "wdth",
    "Optical Size": "opsz",
    "Serif": "SERF",
}

ignore = [
    "Encoding Mac",
    "Encoding",
    "Master Count",
    "Default Weight Vector",
    "Type 1 Unique ID",
    "Menu Name",
    "FOND Name",
    "weight_name",  # Do something with this?
    "width_name",  # Do something with this?
    "weight",  # Do something if single master
    "trademark",
    "Monospaced",
    "Slant Angle",
    "Background Bitmap",
    "Glyph Origin",
    "Glyph Anchors Supplemental",
    "Links",
    "PostScript Info",  # for now; one per master
]
store_in_scratch = [
    "Axis Mappings Count",
]
names = {
    "description": "description",
    "License": "license",
    "License URL": "licenseURL",
    "designer": "designer",
    "designerURL": "designerURL",
    "manufacturer": "manufacturer",
    "manufacturerURL": "manufacturerURL",
    "copyright": "copyright",
    "sgn": "familyName",
    "tfn": "familyName",
    "versionFull": "version",
}

# hhea_line_gap
# hhea_ascender
# hhea_descender
# openTypeFeatures
# OpenType Class (names -> glyphs)


# Axis count
# Axis Name
# Axis Mappings Count (items per axis?)


class FontlabVFB(BaseConvertor):
    suffix = ".vfb"

    def _load(self):
        try:
            from vfbLib.vfb.vfb import Vfb
        except ImportError:
            raise ImportError(
                "This convertor requires the vfbLib package. Install it with `pip install vfbLib`"
            )

        self.vfb = Vfb(self.filename)
        self.vfb.decompile()
        self.current_glyph = None
        self.current_master = None
        scratch = defaultdict(list)
        self.glyph_names = []
        self.metrics = {}
        # Quickly set up GID->name mapping
        for e in self.vfb.entries:
            if e.key == "Glyph":
                self.glyph_names.append(e.decompiled["name"])

        # Now parse the whole thing
        for e in self.vfb.entries:
            name = e.key
            if name is None:
                raise TypeError
            data = e.decompiled
            if data is None:
                continue

            if name in ignore or re.match("^\d+$", name):
                continue
            if name in store_in_scratch:
                scratch[name].append(data)
                continue

            if name == "psn":
                # Postscript name, hey we don't have that.
                pass
            elif name in names:
                if data:
                    setattr(
                        self.font.names, names[name], I18NDictionary.with_default(data)
                    )
            # Axes
            elif name == "Axis Name":
                axis = Axis(name, tags.get(name, name.upper()[:4]))  # Fix up tag!
                self.font.axes.append(axis)
            elif name == "Axis Mappings":
                counts = scratch["Axis Mappings Count"][0]
                for axis, count in zip(self.font.axes, counts):
                    if count > 0:
                        mappings = data[:count]
                        axis.map = [(u, round(d * 1000)) for u, d in mappings]
                        user_coords = [c[0] for c in mappings]
                        axis.min = min(user_coords, default=0)
                        axis.default = axis.minimum
                        axis.max = max(user_coords, default=1000)
                    data = data[10:]
            elif name == "Master Name":
                self.current_master = Master(name=data, id=uuid.uuid1())
                if self.metrics:
                    self.current_master.metrics = self.metrics
                self.font.masters.append(self.current_master)
            elif name == "Master Location":
                _, location = data
                master = self.current_master
                master.location = {}
                for axis, value in zip(self.font.axes, location):
                    master.location[axis.tag] = value
            elif name == "ffn":  # Full family name?
                pass
            elif name == "upm":
                self.font.upm = int(data)
            elif name == "versionMajor":
                self.font.version = (int(data), self.font.version[1])
            elif name == "versionMinor":
                self.font.version = (self.font.version[0], int(data))
            elif name == "vendorID":
                self.font.custom_opentype_values[("OS/2", "achVendID")] = data
            elif name == "hhea_line_gap":
                self.metrics["hheaLineGap"] = int(data)
            elif name == "hhea_ascender":
                self.metrics["hheaAscender"] = int(data)
            elif name == "hhea_descender":
                self.metrics["hheaDescender"] = int(data)
            elif name == "Glyph":
                self.current_glyph = Glyph(name=data["name"])
                self.font.glyphs.append(self.current_glyph)
                self._load_glyph(data)
            elif name == "Glyph GDEF Data":
                for anchor in data.get("anchors", []):
                    self._load_anchor(anchor)
            elif name == "Glyph Guide Properties":
                pass
            elif name == "TrueType Info":
                self._handle_truetype_info(data)
            elif name == "Glyph Unicode":
                self.current_glyph.codepoints = data
            elif name == "Italic Angle":
                # Put in master
                pass
            else:
                print(name, data)

        # Sort the glyphs by Unicode ID and name, because there's no
        # other order in there.
        glyphorder = list(self.font.glyphs.items())
        glyphorder.sort(key=lambda k: (k[1].codepoints, k[0]))
        newlist = GlyphList()
        for g in glyphorder:
            newlist.append(g[1])
        self.font.glyphs = newlist

        # Try and fill in metrics we don't have
        for master in self.font.masters:
            if "ascender" not in master.metrics and "hheaAscender" in master.metrics:
                master.metrics["ascender"] = master.metrics["hheaAscender"]
            if "descender" not in master.metrics and "hheaDescender" in master.metrics:
                master.metrics["descender"] = master.metrics["hheaDescender"]
        return self.font

    def _load_glyph(self, data):
        metrics = data["metrics"]
        masters = self.font.masters
        layers = [
            Layer(width=metric[0], height=metric[1], id=master.id)
            for master, metric in zip(masters, metrics)
        ]
        pens = [layer.getPen() for layer in layers]
        for node in data["nodes"]:
            for pen, point in zip(pens, node["points"]):
                cmd = getattr(pen, node["type"] + "To")
                if node["type"] == "move" and pen.contour:
                    pen.closePath()
                if node["type"] == "curve":
                    pen.curveTo(point[1], point[2], point[0])
                else:
                    cmd(*point)
        for pen in pens:
            if pen.contour:
                pen.closePath()
        for component in data.get("components", []):
            for ix, pen in enumerate(pens):
                pen.addComponent(
                    self.glyph_names[component["gid"]],
                    Transform(
                        dx=component["offsetX"][ix],
                        dy=component["offsetY"][ix],
                        xx=component["scaleX"][ix],
                        yy=component["scaleY"][ix],
                    ),
                )
        for layer in layers:
            self.current_glyph.layers.append(layer)

    def _load_anchor(self, data):
        layer = self.current_glyph.layers[-1]
        layer.anchors.append(
            Anchor(
                name=data["name"],
                x=data["x"],
                y=data["y"],
            )
        )

    def _handle_truetype_info(self, data):
        for element in data:
            pass



import logging
import re

from babelfont import Font

logger = logging.getLogger(__name__)


class NibbleParser:
    def __init__(self, target):
        self.string = target

    def __call__(self, regex):
        if m := re.match(regex, self.string):
            self.string = re.sub(regex, "", self.string)
            return m
        return None


def translate_glyphs3_fea(font: Font, args=None):
    font.features.features = [
        (key, _translate(code, {"index": 0, "font": font, "tag": key}))
        for key, code in font.features.features
    ]


def _translate(code, state):
    return re.sub(
        "#ifdef VARIABLE(.*)#endif",
        lambda m: __translate(m, state),
        code,
        flags=re.DOTALL,
    ).replace("{;", "{")


def __translate(match, state):
    newstatements = []

    def name():
        return f"__condition_{state['tag']}_{state['index']}"

    axis_map = {a.tag: a for a in state["font"].axes}

    for statement in match.group(1).split(";"):
        p = NibbleParser(statement)
        if p(r"^\s*condition\s+"):
            axes = []
            while axis := p(
                r"""(?x)
                    (?:([\d.]+)\s*<)? # Axis minimum value
                    \s*(\w+)\s*    # Axis tag
                    (?:<\s*([\d.]+))? # Axis maximum value
                    ,?\s*             # Optional comma
            """
            ):
                min_v, tag, max_v = axis.groups()
                if tag not in axis_map:
                    raise ValueError(f"Axis {tag} not found in condition statement")
                if min_v is None:
                    min_v = axis_map[tag].minimum
                if max_v is None:
                    max_v = axis_map[tag].maximum
                axes.append(f"{tag} {min_v} {max_v}")
            if not axes:
                raise ValueError("No axis definition found for condition")

            if not newstatements:
                newstatements.append("")
            newstatements.append(f"}} {state['tag']}")
            state["index"] += 1
            newstatements.append(f"conditionset {name()} {{")
            newstatements.extend(axes)
            newstatements.append(f"}} {name()}")
            newstatements.append(f"variation {state['tag']} {name()} {{")
        else:
            newstatements.append(statement)
    return ";".join(newstatements)



import uuid
from datetime import datetime
from itertools import chain
from typing import Dict

from fontFeatures import Attachment
from fontFeatures.ttLib import unparse
from fontTools.fontBuilder import FontBuilder
from fontTools.misc.fixedTools import otRound
from fontTools.misc.timeTools import epoch_diff, timestampSinceEpoch
from fontTools.pens.recordingPen import RecordingPen
from fontTools.pens.ttGlyphPen import TTGlyphPen
from fontTools.ttLib import TTFont
from fontTools.ttLib.tables._g_l_y_f import GlyphCoordinates
from fontTools.ttLib.tables.O_S_2f_2 import Panose
from fontTools.ttLib.tables.TupleVariation import TupleVariation
from fontTools.ttLib.ttGlyphSet import _TTGlyph
from fontTools.varLib.iup import iup_delta_optimize

from babelfont import (
    Anchor,
    Axis,
    Features,
    Glyph,
    Instance,
    Layer,
    Master,
    Node,
    Shape,
)
from babelfont.convertors import BaseConvertor
from babelfont.fontFilters.featureWriters import build_all_features


def _categorize_glyph(font, glyphname):
    if "GDEF" not in font:
        return None
    if not font["GDEF"].table.GlyphClassDef:
        return None
    classdefs = font["GDEF"].table.GlyphClassDef.classDefs
    if glyphname not in classdefs:
        return None
    if classdefs[glyphname] == 1:
        return "base"
    if classdefs[glyphname] == 2:
        return "ligature"
    if classdefs[glyphname] == 3:
        return "mark"
    if classdefs[glyphname] == 4:
        return "component"


def compile_panose(data):
    panose = Panose()
    panose.bFamilyType = data[0]
    panose.bSerifStyle = data[1]
    panose.bWeight = data[2]
    panose.bProportion = data[3]
    panose.bContrast = data[4]
    panose.bStrokeVariation = data[5]
    panose.bArmStyle = data[6]
    panose.bLetterForm = data[7]
    panose.bMidline = data[8]
    panose.bXHeight = data[9]
    return panose


class TrueType(BaseConvertor):
    suffix = ".ttf"

    def _decompose_mixed_layer(self, layer, exportable):
        if (layer.paths and layer.components) or any(
            c.ref not in exportable for c in layer.components
        ):
            layer.decompose()

    def _load(self):
        self.tt = TTFont(self.filename)
        self._load_fvar()
        self._load_head()
        self._load_masters()
        self._load_names()
        self._load_glyphs()
        self._load_features()
        return self.font

    def _load_fvar(self):
        avar = self.tt.get("avar")
        if "fvar" in self.tt:
            for axis in self.tt["fvar"].axes:
                name = self.tt["name"].getDebugName(axis.axisNameID)  # XXX multilingual
                bb_axis = Axis(
                    tag=axis.axisTag,
                    min=axis.minValue,
                    max=axis.maxValue,
                    default=axis.defaultValue,
                    name=name,
                )
                self.font.axes.append(bb_axis)
                if avar:
                    segs = avar.segments[axis.axisTag]
                    mapping = {
                        bb_axis.denormalize_value(k): bb_axis.denormalize_value(v)
                        for k, v in segs.items()
                    }
                    bb_axis.map = mapping
            for instance in self.tt["fvar"].instances:
                self.font.instances.append(
                    Instance(
                        name=self.tt["name"].getDebugName(instance.subfamilyNameID),
                        location=instance.coordinates,
                    )
                )

    def _load_masters(self):
        m = Master(location={}, name="Default", id=str(uuid.uuid1()))
        # Metrics
        m.metrics = {
            "xHeight": (
                self.tt["OS/2"].sxHeight
                if hasattr(self.tt["OS/2"], "sxHeight")
                else None
            ),
            "capHeight": (
                self.tt["OS/2"].sCapHeight
                if hasattr(self.tt["OS/2"], "capHeight")
                else None
            ),
            "ascender": self.tt["hhea"].ascender,
            "descender": self.tt["hhea"].descender,
        }
        m.font = self.font
        self.font.masters = [m]
        if "fvar" in self.tt:
            m.location = {axis.tag: axis.default for axis in self.font.axes}
            all_masters = [
                frozenset(x.axes.items())
                for x in chain(*self.tt["gvar"].variations.values())
            ]
            all_masters = [{k: v[1] for k, v in dict(m1).items()} for m1 in all_masters]
            # Now denormalize.
            # XXX
            pass

    def _load_head(self):
        head = self.tt["head"]
        self.font.upm = head.unitsPerEm
        minor = head.fontRevision % 1
        while minor - int(minor) > 1e-4:
            minor *= 10
        self.font.version = (int(head.fontRevision), int(minor))
        self.font.date = datetime.fromtimestamp(self.tt["head"].created + epoch_diff)

    def _load_names(self):
        names = self.tt["name"]
        # XXX

    def _load_glyphs(self):
        mapping = self.tt["cmap"].buildReversed()
        glyphs_dict = {}
        for glyph in self.tt.getGlyphOrder():
            category = _categorize_glyph(self.tt, glyph) or "base"
            glyphs_dict[glyph] = Glyph(
                name=glyph, codepoints=list(mapping.get(glyph, [])), category=category
            )
            self.font.glyphs.append(glyphs_dict[glyph])
            glyphs_dict[glyph].layers = self._load_layers(glyph, glyphs_dict[glyph])
        return glyphs_dict

    def _load_layers(self, glyphname, glyph):
        ttglyph = self.tt.getGlyphSet()[glyphname]  # _TTGlyphGlyf object
        width = self.tt["hmtx"][glyphname][0]
        # leftMargin = self.tt["hmtx"][g][1]
        layer = Layer(width=width, id=str(uuid.uuid1()))
        layer._master = self.font.masters[0].id
        layer._font = self.font
        layer._glyph = glyph
        ttglyph.draw(layer.getPen())
        return [layer]

    SAVE_FILTERS = [
        "renameGlyphs:production=True",
        "decomposeMixedGlyphs",
        "dropUnexportedGlyphs",
        "zeroMarkWidths",
        "cubicToQuadratic",
        "fillOpentypeValues",
    ]

    def _save(self):
        f = self.font
        fb = FontBuilder(f.upm, isTTF=True)
        fb.setupGlyphOrder(list(f.glyphs.keys()))
        fb.setupCharacterMap(f.unicode_map)

        metrics = {}
        for g in f.glyphs.keys():
            layer = f.default_master.get_glyph_layer(g)
            metrics[g] = (layer.width or 0, layer.lsb or 0)

        fb.setupHorizontalMetrics(metrics)

        ttglyphsets: Dict[str, Dict[str, _TTGlyph]] = {m.id: {} for m in f.masters}
        # We need to do this in order of single components first
        done = set()

        def convert_glyph(g):
            if g in done:
                return
            for m in f.masters:
                layer = m.get_glyph_layer(g)
                if layer:
                    for c in layer.components:
                        convert_glyph(c.ref)
                    pen = TTGlyphPen(ttglyphsets[m.id])
                    layer.draw(pen)

                    ttglyphsets[m.id][g] = pen.glyph()
            done.add(g)

        for g in f.glyphs.keys():
            convert_glyph(g)

        try:
            fb.setupGlyf(ttglyphsets[f.default_master.id])
        except ValueError:
            fb.font["head"].glyphDataFormat = 1
            self.logger.warning(
                "Using Boring Expansion, eh? Setting glyf data format to 1"
            )
            fb.setupGlyf(ttglyphsets[f.default_master.id])
        fb.setupHorizontalHeader()
        f.names.typographicSubfamily = f.default_master.name
        f.names.typographicFamily = f.names.familyName
        fb.setupNameTable(f.names.as_nametable_dict(), mac=False)
        fb.setupOS2()

        if f.axes:
            model = f.variation_model()
            axis_map = {}
            variations = {}
            self.logger.info("Calculating gvar table")
            for g in f.glyphs.keys():
                variations[g] = self.calculate_a_gvar(f, model, g, ttglyphsets)

            for ax in f.axes:
                ax.name = ax.name.as_fonttools_dict
                axis_map[ax.tag] = ax
            for instance in f.instances:
                instance.location = {
                    k: axis_map[k].map_backward(v) for k, v in instance.location.items()
                }
            fb.setupFvar(f.axes, f.instances)

            fb.setupGvar(variations)
            fb.setupAvar(f.axes)
        build_all_features(f, fb.font)
        fb.setupPost()
        # Set OS/2 version to 4 for ufo2ft compatibility
        fb.font["OS/2"].version = 4
        fb.font["maxp"].maxZones = 1

        for (table, field), value in f.custom_opentype_values.items():
            if field == "panose":
                value = compile_panose(value)
            setattr(fb.font[table], field, value)
            self.logger.debug("Setting %s.%s to %s", table, field, value)
        fb.font.save(self.filename)

    def calculate_a_gvar(
        self, f, model, g, ttglyphsets: Dict[str, Dict[str, _TTGlyph]]
    ):
        if g not in ttglyphsets[f.default_master.id]:
            return None
        default_g = ttglyphsets[f.default_master.id][g]
        all_coords = []
        layers = [m.get_glyph_layer(g) for m in f.masters]
        submodel, _ = model.getSubModel(layers)
        for m in f.masters:
            layer = m.get_glyph_layer(g)
            if not layer:
                continue
            masterglyph = ttglyphsets[m.id][g]
            basecoords = GlyphCoordinates(masterglyph.coordinates)
            if masterglyph.isComposite():
                component_point = GlyphCoordinates(
                    [
                        (otRound(layer_comp.pos[0]), otRound(layer_comp.pos[1]))
                        for layer_comp in layer.components
                    ]
                )
                basecoords.extend(component_point)
            phantomcoords = GlyphCoordinates(
                [(0, 0), (otRound(layer.width), 0), (0, 0), (0, 0)]
            )
            basecoords.extend(phantomcoords)
            all_coords.append(basecoords)
        for ix, c in enumerate(all_coords):
            all_ok = True
            if len(c) != len(all_coords[0]):
                print("Incompatible master %i in glyph %s" % (ix, g))
                all_ok = False
            if not all_ok:
                return []
        deltas = submodel.getDeltas(all_coords)
        gvar_entry = []
        if default_g.isComposite():
            endPts = list(range(len(default_g.components)))
        else:
            endPts = default_g.endPtsOfContours

        for delta, sup in zip(deltas, submodel.supports):
            if not sup:
                continue
            var = TupleVariation(sup, round(delta))
            # This assumes we do the default master first, which may not be true
            delta_opt = iup_delta_optimize(
                round(delta), round(deltas[0]), endPts, tolerance=0.5
            )
            if None in delta_opt:
                var = TupleVariation(sup, delta_opt)
            gvar_entry.append(var)
        return gvar_entry

    def _load_features(self):
        features = unparse(self.tt)
        # Load anchors
        for routine in features.routines:
            for rule in routine.rules:
                if isinstance(rule, Attachment):
                    for glyphname, pos in rule.bases.items():
                        self._add_anchor(glyphname, pos, rule.base_name)
                    for glyphname, pos in rule.marks.items():
                        self._add_anchor(glyphname, pos, rule.mark_name)

        self.font.features = Features()
        for feature, routines in features.features.items():
            self.font.features.features.append(
                (str(feature), "\n".join(x.asFea() for x in routines))
            )
        for routine in features.routines:
            self.font.features.prefixes[routine.name] = routine.asFea()
        self.font.features.classes = features.namedClasses

    def _add_anchor(self, glyphname, pos, name):
        # Would be nice if this was variable.
        layer = self.font.default_master.get_glyph_layer(glyphname)
        layer.anchors.append(Anchor(name=name, x=pos[0], y=pos[1]))

    # import numpy as np
    # def calculate_a_gvar(self, f, model, g, default_width):
    #     if not g in f.default_master.ttglyphset._glyphs:
    #         return None

    #     all_coords = []
    #     master_ix = f.masters.index(f.default_master)

    #     for m in f.masters:
    #         coords = list(m.ttglyphset._glyphs[g].coordinates)
    #         layer = m.get_glyph_layer(g)
    #         if m.ttglyphset._glyphs[g].isComposite():
    #             coords.extend([c.pos for c in layer.components])
    #         coords.extend( [ (0,0), (layer.width,0), (0,0), (0,0) ] )
    #         all_coords.append(np.array(coords))
    #     stacked = np.array(all_coords)
    #     defaults = stacked[np.newaxis,master_ix,:,:].repeat(len(f.masters),axis=0)
    #     base_deltas = stacked-defaults
    #     x_deltas = np.apply_along_axis(model.getDeltas, 1, base_deltas[:,:,0].transpose())
    #     y_deltas = np.apply_along_axis(model.getDeltas, 1, base_deltas[:,:,1].transpose())
    #     alldeltas = np.array([x_deltas, y_deltas]).transpose()
    #     gvar_entry = []
    #     for deltaset, sup in zip(alldeltas, model.supports):
    #         gvar_entry.append(TupleVariation(sup, list(map(tuple, deltaset))))
    #     return gvar_entry


class OpenType(TrueType):
    suffix = ".otf"

    def _load_layers(self, g, _glyph):
        ttglyph = self.tt.getGlyphSet()[g]
        width = self.tt["hmtx"][g][0]
        layer = Layer(width=width, id=self.font.masters[0].id)
        layer._master = self.font.masters[0].id
        layer._font = self.font
        pen = RecordingPen()
        ttglyph.draw(pen)
        contours = pen.value
        lastcontour = []
        startPt = (0, 0)
        lastPt = (0, 0)
        index = 0
        for operation, segment in contours:
            if operation == "moveTo":
                startPt = segment[0]
            elif operation == "closePath":
                if startPt != lastPt:
                    lastcontour.append(Node(x=startPt[0], y=startPt[1], type="l"))
                contour = Shape()
                contour.nodes = lastcontour
                layer.shapes.append(contour)
                lastcontour = []
            elif operation == "curveTo":
                lastcontour.append(Node(x=segment[0][0], y=segment[0][1], type="o"))
                lastcontour.append(Node(x=segment[1][0], y=segment[1][1], type="o"))
                lastcontour.append(Node(x=segment[2][0], y=segment[2][1], type="c"))
                lastPt = segment[2]
            elif operation == "lineTo":
                lastcontour.append(Node(x=segment[0][0], y=segment[0][1], type="l"))
                lastPt = segment[0]
            elif operation == "qCurveTo":
                lastcontour.append(Node(x=segment[0][0], y=segment[0][1], type="o"))
                lastcontour.append(Node(x=segment[1][0], y=segment[1][1], type="q"))

        return [layer]



import argparse
import logging
import sys

from babelfont.convertors import Convert
from babelfont.convertors.truetype import TrueType
from babelfont.fontFilters import FILTERS, parse_filter

LOG_FORMAT = "%(message)s"

logger = logging.getLogger(__name__)


def main():
    parser = argparse.ArgumentParser(
        prog="babelfont", description="Convert between font formats"
    )
    parser.add_argument(
        "--log-level",
        "-l",
        help="Log level",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        default="INFO",
    )
    parser.add_argument(
        "--filter",
        "-f",
        help="Filter to apply",
        action="append",
        # choices=FILTERS.keys(),
    )
    parser.add_argument(
        "--disable-filter",
        help="Filter to disable",
        action="append",
        choices=FILTERS.keys(),
    )
    parser.add_argument("input", metavar="IN", help="Input file")
    parser.add_argument("output", metavar="OUT", help="Output file")
    args = parser.parse_args()

    try:
        from rich.logging import RichHandler

        handlers = [RichHandler()]
    except ImportError:
        handlers = [logging.StreamHandler()]

    logging.basicConfig(
        level=args.log_level, format=LOG_FORMAT, datefmt="[%X]", handlers=handlers
    )

    input_job = Convert(args.input)
    convertor_in = input_job.load_convertor()
    if not convertor_in:
        sys.exit(1)

    try:
        logger.info("Reading %s", args.input)
        font = convertor_in.load(input_job, filters=False)
    except Exception as e:
        if args.log_level == "DEBUG":
            raise e
        logger.error("Couldn't read %s: %s", args.input, e)
        sys.exit(1)

    output_job = Convert(args.output)
    convertor_out = output_job.save_convertor()
    if not convertor_out:
        sys.exit(1)

    filters = convertor_in.LOAD_FILTERS
    filters += args.filter or []
    if isinstance(convertor_out, TrueType):
        filters += convertor_in.COMPILE_FILTERS
    filters += convertor_out.SAVE_FILTERS
    if args.disable_filter:
        filters = [f for f in filters if f not in args.disable_filter]

    for filter in filters:
        fltr, filterargs = parse_filter(filter)
        fltr(font, filterargs)

    try:
        logger.info("Saving %s", args.output)
        convertor_out.save(font, output_job)
    except Exception as e:
        logger.error("Couldn't write %s: %s", args.output, e)
        if args.log_level == "DEBUG":
            raise e

        sys.exit(1)

    sys.exit(0)


if __name__ == "__main__":
    sys.exit(main())



from .BaseObject import BaseObject, I18NDictionary
from dataclasses import dataclass, asdict

OPENTYPE_NAMES = [
    "copyright",
    "familyName",
    "preferredSubfamilyName",
    "uniqueID",
    "fullName",  # XXX?
    "version",
    "postscriptName",
    "trademark",
    "manufacturer",
    "designer",
    "description",
    "manufacturerURL",
    "designerURL",
    "license",
    "licenseURL",
    "reserved",
    "typographicFamily",
    "typographicSubfamily",
    "compatibleFullName",
    "sampleText",
    "postscriptCIDname",  # XXX?
    "WWSFamilyName",
    "WWSSubfamilyName",
]


@dataclass
class Names(BaseObject):
    """A table of global, localizable names for the font."""

    familyName: I18NDictionary = None
    designer: I18NDictionary = None
    designerURL: I18NDictionary = None
    manufacturer: I18NDictionary = None
    manufacturerURL: I18NDictionary = None
    license: I18NDictionary = None
    licenseURL: I18NDictionary = None
    version: I18NDictionary = None
    uniqueID: I18NDictionary = None
    postscriptName: I18NDictionary = None
    fullName: I18NDictionary = None
    description: I18NDictionary = None
    typographicFamily: I18NDictionary = None
    typographicSubfamily: I18NDictionary = None
    compatibleFullName: I18NDictionary = None
    sampleText: I18NDictionary = None
    WWSFamilyName: I18NDictionary = None
    WWSSubfamilyName: I18NDictionary = None
    copyright: I18NDictionary = None
    styleMapFamilyName: I18NDictionary = None
    styleMapStyleName: I18NDictionary = None
    styleName: I18NDictionary = None
    trademark: I18NDictionary = None
    preferredSubfamilyName: I18NDictionary = None

    def __post_init__(self):
        for k in self.__dataclass_fields__.keys():
            if not getattr(self, k):
                setattr(self, k, I18NDictionary())

    def as_nametable_dict(self):
        rv = {}
        ft_names = {
            "manufacturerURL": "vendorURL",
            "license": "licenseDescription",
            "licenseURL": "licenseInfoURL",
        }
        for k, v in asdict(self).items():
            if not v:
                continue
            rv[ft_names.get(k, k)] = v.default_or_dict()
        return rv



from collections import defaultdict
import logging

from fontTools.feaLib import ast
from fontTools.misc.visitor import Visitor

from babelfont.Font import Font

logger = logging.getLogger(__name__)


def drop_unexported_glyphs(font: Font, _args=None):
    logger.info("Dropping unexported glyphs")
    unexported = set(glyph.name for glyph in font.glyphs if not glyph.exported)
    # Safety check one: look in components:
    appearances = defaultdict(set)
    for glyph in font.glyphs:
        for layer in glyph.layers:
            for c in layer.components:
                if c.ref in unexported:
                    appearances[c.ref].add(glyph.name)
    if appearances:
        logger.warning(
            "Unexported glyphs are used in components, use decomposeMixed first:"
        )
        for glyph, refs in appearances.items():
            logger.warning(f"  {glyph} -> {', '.join(refs)}")
        logger.warning("Will not drop these glyphs.")
        unexported -= set(appearances.keys())
    # Safety check two: look in features
    parsed_features = font.features.as_ast(font)

    for featurename, parsed in parsed_features["features"]:
        visitor = FeaAppearsVisitor()
        visitor.visit(parsed)
        for glyph in visitor.appearances:
            if glyph in unexported:
                logger.warning(
                    f"Unexported glyph {glyph} is used in feature {featurename}, will not drop it."
                )
                unexported.remove(glyph)
    for classname, glyphs in font.features.classes.items():
        for glyph in glyphs:
            if glyph in unexported:
                logger.warning(
                    f"Unexported glyph {glyph} is used in class @{classname}, will not drop it."
                )
                unexported.remove(glyph)
    for prefix, parsed in parsed_features["prefixes"].items():
        visitor = FeaAppearsVisitor()
        visitor.visit(parsed)
        for glyph in visitor.appearances:
            if glyph in unexported:
                logger.warning(
                    f"Unexported glyph {glyph} is used in prefix {prefix}, will not drop it."
                )
                unexported.remove(glyph)

    # Now we are good to go
    for glyph in unexported:
        font.glyphs.pop(glyph)


class FeaAppearsVisitor(Visitor):
    def __init__(self):
        self.appearances = set()


@FeaAppearsVisitor.register(ast.GlyphName)
def visit(visitor, gn, *args, **kwargs):
    visitor.appearances.add(gn.glyph)
    return False


@FeaAppearsVisitor.register(ast.MarkClassDefinition)
def visit(visitor, mcd, *args, **kwargs):
    visitor.visitAttr(mcd, "glyphs", mcd.glyphs, *args, **kwargs)
    return False



import re
import shlex
import uuid
import logging
from pathlib import Path

from fontTools.misc.timeTools import epoch_diff

from babelfont.convertors import BaseConvertor
from babelfont.Names import OPENTYPE_NAMES

from babelfont import (
    Anchor,
    Axis,
    Master,
    Instance,
    Glyph,
    Layer,
    Shape,
    Transform,
)

NUMBER = r"(-?\d+(?:\.\d+)?(?:e-?\d+)?)"
QUOTED_STRING = r'"([^"]+)"'
UNQUOTED_STRING = r"(\S+)"
MOVE_RE = re.compile(rf"^\s*{NUMBER}\s+{NUMBER}\s+m")
CURVE_RE = re.compile(
    rf"^\s*{NUMBER}\s+{NUMBER}\s+{NUMBER}\s+{NUMBER}\s+{NUMBER}\s+{NUMBER} c"
)
LINE_RE = re.compile(rf"^\s*{NUMBER}\s+{NUMBER}\s+l")

log = logging.getLogger(__name__)


class FontForgeSFDIR(BaseConvertor):
    suffix = ".sfdir"
    IGNORE = [
        "SplineFontDB",
        "UComments",
        "TeXData",
        "FontName",
        "DefaultBaseFilename",
        "sfntRevision",
        "InvalidEm",
        "XUID",
        "OS2Version",
        "PfmFamily",
        "AntiAlias",
        "FitToEm",
        "DisplaySize",
        "Validated",
        "Flags",
        "Lookup",
        "GlifName",
        "NameList",
    ]

    @classmethod
    def can_save(cls, convertor, **kwargs):
        return False

    def _setupfont(self):
        font = self.font
        font.upm = 1024
        # I'm going to assume a single master font
        font.masters = [
            Master(
                name="Regular", id=str(uuid.uuid4()), font=font, location={"wght": 400}
            )
        ]
        font.axes = [Axis(name="Weight", tag="wght", min=400, max=400, default=400)]
        font.instances = [
            Instance(name="Regular", styleName="Regular", location={"wght": 400})
        ]

    def _load(self):
        self._setupfont()
        self.current_glyph = None
        self.info = open(Path(self.filename) / "font.props").read().splitlines()
        self._load_line("info")
        # Probe once for glyphorder
        self.glyph_order = {}
        for fontfile in Path(self.filename).glob("*.glyph"):
            self._probe_glyphorder(open(fontfile).readlines())
        for gid in sorted(self.glyph_order.keys()):
            self.font.glyphs.append(Glyph(name=self.glyph_order[gid]))

        for fontfile in Path(self.filename).glob("*.glyph"):
            self.info = open(fontfile).read().splitlines()
            self._load_line("glyph")

        return self.font

    def _probe_glyphorder(self, lines):
        glyphname = ""
        for line in lines:
            if m := re.match(r"StartChar: (.*)", line):
                glyphname = m.group(1)
                continue
            if m := re.match(r"Encoding: (\d+) (-?\d+) (\d+)", line):
                gid = int(m.group(3))
                if not glyphname:
                    raise ValueError(f"Bad file - no StartChar")
                self.glyph_order[gid] = glyphname

    def _load_line(self, mode):
        while self.info:
            line = self.info.pop(0)
            if line == "EndSplineFont":
                break
            if line.startswith("BeginChars") and mode == "info":
                break
            if line == "EndChar" and mode == "glyph":
                break
            if not line:
                continue
            if ": " not in line:
                if hasattr(self, "_handle_" + line):
                    getattr(self, "_handle_" + line)()
                    continue
                else:
                    raise ValueError("Bad font.props file - " + line)
            key, value = line.split(": ", 1)
            if key in self.IGNORE:
                continue
            if hasattr(self, "_handle_" + key):
                getattr(self, "_handle_" + key)(value)
            else:
                log.debug("Unknown info key %s", key)
                pass

    def _handle_FamilyName(self, value):
        self.font.names.familyName.set_default(value)

    def _handle_FullName(self, value):
        self.font.names.compatibleFullName.set_default(value)

    def _handle_Copyright(self, value):
        self.font.names.copyright.set_default(value)

    def _handle_Comments(self, value):
        self.font.notes = value

    def _handle_Version(self, value):
        version = [int(x) for x in value.split(".")]
        if len(version) == 2:
            self.font.version = tuple(version)
        elif len(version) == 1:
            self.font.version = (version[0], 0)
        else:
            version2 = version[1] * 100 + version[2]
            while version2 > 999:
                version2 /= 10
            self.font.version = (version[0], version2)

    def _handle_Weight(self, value):
        self.font.names.typographicSubfamily.set_default(value)

    def _handle_Fontlog(self, value):
        self.font.names.description.set_default(value)

    def _handle_ItalicAngle(self, value):
        self.font.masters[0].metrics["italicAngle"] = float(value)

    def _handle_UnderlinePosition(self, value):
        self.font.masters[0].metrics["underlinePosition"] = int(value)

    def _handle_UnderlineWidth(self, value):
        self.font.masters[0].metrics["underlineThickness"] = int(value)

    def _handle_OS2TypoAscent(self, value):
        self.font.masters[0].metrics["typoAscender"] = int(value)

    def _handle_OS2TypoDescent(self, value):
        self.font.masters[0].metrics["typoDescender"] = int(value)

    def _handle_OS2TypoLinegap(self, value):
        self.font.masters[0].metrics["typoLineGap"] = int(value)

    def _handle_OS2WinAscent(self, value):
        self.font.masters[0].metrics["winAscent"] = int(value)

    def _handle_OS2WinDescent(self, value):
        self.font.masters[0].metrics["winDescent"] = int(value)

    def _handle_OS2Vendor(self, value):
        value = value.replace("'", "")
        self.font.custom_opentype_values[("OS/2", "achVendID")] = value

    def _handle_HheadAscent(self, value):
        self.font.masters[0].metrics["hheaAscender"] = int(value)

    def _handle_HheadDescent(self, value):
        self.font.masters[0].metrics["hheaDescender"] = int(value)

    def _handle_LineGap(self, value):
        self.font.masters[0].metrics["hheaLineGap"] = int(value)

    def _handle_Ascent(self, value):
        self.font.masters[0].metrics["ascender"] = int(value)

    def _handle_Descent(self, value):
        self.font.masters[0].metrics["descender"] = -int(value)

    def _handle_OS2SubXSize(self, value):
        self.font.masters[0].metrics["subscriptXSize"] = int(value)

    def _handle_OS2SubYSize(self, value):
        self.font.masters[0].metrics["subscriptYSize"] = int(value)

    def _handle_OS2SubXOff(self, value):
        self.font.masters[0].metrics["subscriptXOffset"] = int(value)

    def _handle_OS2SubYOff(self, value):
        self.font.masters[0].metrics["subscriptYOffset"] = int(value)

    def _handle_OS2SupXSize(self, value):
        self.font.masters[0].metrics["superscriptXSize"] = int(value)

    def _handle_OS2SupYSize(self, value):
        self.font.masters[0].metrics["superscriptYSize"] = int(value)

    def _handle_OS2SupXOff(self, value):
        self.font.masters[0].metrics["superscriptXOffset"] = int(value)

    def _handle_OS2SupYOff(self, value):
        self.font.masters[0].metrics["superscriptYOffset"] = int(value)

    def _handle_OS2StrikeYSize(self, value):
        self.font.masters[0].metrics["strikeoutSize"] = int(value)

    def _handle_OS2StrikeYPos(self, value):
        self.font.masters[0].metrics["strikeoutPosition"] = int(value)

    def _handle_OS2CapHeight(self, value):
        self.font.masters[0].metrics["capHeight"] = int(value)

    def _handle_OS2XHeight(self, value):
        self.font.masters[0].metrics["xHeight"] = int(value)

    def _handle_HheadAscent(self, value):
        self.font.masters[0].metrics["hheaAscender"] = int(value)

    def _handle_LayerCount(self, value):
        # Ignore layers for now
        pass

    def _handle_Layer(self, value):
        # Ignore layers for now
        pass

    def _handle_FSType(self, value):
        self.font.custom_opentype_values[("OS/2", "fsType")] = int(value)

    def _handle_OS2_WeightWidthSlopeOnly(self, value):
        pass

    def _handle_TTFWeight(self, value):
        self.font.custom_opentype_values[("OS/2", "usWeightClass")] = int(value)

    def _handle_TTFWidth(self, value):
        self.font.custom_opentype_values[("OS/2", "usWidthClass")] = int(value)

    def _handle_Panose(self, value):
        bits = [int(v) for v in value.split()]
        self.font.custom_opentype_values[("OS/2", "bPanose")] = bits

    def _handle_OS2_UseTypoMetrics(self, value):
        if not value:
            return
        if ("OS/2", "fsSelection") in self.font.custom_opentype_values:
            self.font.custom_opentype_values[("OS/2", "fsSelection")] |= 1 << 7
        else:
            self.font.custom_opentype_values[("OS/2", "fsSelection")] = 1 << 7

    def _handle_CreationTime(self, value):
        self.font.custom_opentype_values[("head", "created")] = int(value) - epoch_diff

    def _handle_ModificationTime(self, value):
        self.font.custom_opentype_values[("head", "modified")] = int(value) - epoch_diff

    def _handle_ShortTable(self, value):
        # Ignore short table for now
        while self.info:
            line = self.info.pop(0)
            if line.startswith("EndShort"):
                break

    def _handle_TtTable(self, value):
        while self.info:
            line = self.info.pop(0)
            if line.startswith("EndTTInstrs"):
                break

    def _handle_BeginPrivate(self, value):
        while self.info:
            line = self.info.pop(0)
            if line.startswith("EndPrivate"):
                break

    def _handle_Grid(self):
        grid = Layer(name="grid", id="grid", _master="default", _font=self.font)
        self._expect_splineset(grid.getPen())

    def _handle_StartChar(self, value):
        self.current_glyph = self.font.glyphs[value]
        self.current_glyph.layers = [
            Layer(
                name="default",
                id=self.font.masters[0].id,
                _master=self.font.masters[0].id,
                _font=self.font,
                _glyph=self.current_glyph,
            )
        ]

    def _handle_Encoding(self, value):
        if not self.current_glyph:
            return
        _old, unicode, gid = value.split()
        if unicode != "-1":
            self.current_glyph.codepoints.append(int(unicode))

    def _handle_AltUni2(self, value):
        for alternate in value.split():
            codepoint, vs, _reserved = alternate.split(".")
            if vs != "ffffffff":
                log.warning(
                    "Alternate codepoint with variation selector not current supported in %s",
                    self.current_glyph.name,
                )
            self.current_glyph.codepoints.append(int(codepoint, 16))

    def _handle_Width(self, value):
        self.current_glyph.layers[0].width = int(value)

    def _handle_VWidth(self, value):
        self.current_glyph.layers[0].height = int(value)

    def _handle_GlyphClass(self, value):
        pass

    def _handle_Back(self):
        background = Layer(
            name="background",
            id=str(uuid.uuid1()),
            _master=self.font.masters[0].id,
            _font=self.font,
            _glyph=self.current_glyph,
            isBackground=True,
            width=0,
        )
        # self.current_glyph.layers[0].background = background.id
        # self.current_glyph.layers.append(background)
        self._expect_splineset(background.getPen())

    def _handle_Fore(self):
        self._expect_splineset(self.current_glyph.layers[0].getPen())

    def _handle_EndChars(self):
        pass

    def _handle_Refer(self, value: str):
        gid, unicode, style, xx, xy, yx, yy, dx, dy, ttflag = value.split(None, 10)
        component = Shape(
            ref=self.glyph_order[int(gid)],
            transform=Transform(
                float(xx), float(xy), float(yx), float(yy), float(dx), float(dy)
            ),
        )
        self.current_glyph.layers[0].shapes.append(component)

    def _expect_splineset(self, pen):
        paths = False
        while self.info:
            line = self.info.pop(0)
            if line == "SplineSet":
                continue
            if line.startswith("EndSplineSet"):
                break
            if m := re.match(MOVE_RE, line):
                if paths:
                    pen.closePath()
                pen.moveTo((float(m.group(1)), float(m.group(2))))
            elif m := re.match(CURVE_RE, line):
                if (float(m.group(1)), float(m.group(2))) == (
                    float(m.group(3)),
                    float(m.group(4)),
                ):
                    # Quadratic in disguise
                    pen.qCurveTo(
                        (float(m.group(1)), float(m.group(2))),
                        (float(m.group(5)), float(m.group(6))),
                    )
                else:
                    pen.curveTo(
                        (float(m.group(1)), float(m.group(2))),
                        (float(m.group(3)), float(m.group(4))),
                        (float(m.group(5)), float(m.group(6))),
                    )
                paths = True
            elif m := re.match(LINE_RE, line):
                pen.lineTo((float(m.group(1)), float(m.group(2))))
                paths = True
            else:
                # Put back
                self.info.insert(0, line)
                break
        if paths:
            pen.closePath()
        return

    def _handle_AnchorPoint(self, value: str):
        name, x, y, typ, ligcomp = re.match(
            rf"^\s*{QUOTED_STRING}\s+{NUMBER}\s+{NUMBER}\s+{UNQUOTED_STRING}\s+{NUMBER}",
            value,
        ).groups()
        name = name.replace(" ", "")
        if typ == "ligature":
            name += "_%s" % ligcomp
        if typ == "mark":
            name = "_" + name
        self.current_glyph.layers[0].anchors.append(
            Anchor(
                name=name,
                x=float(x),
                y=float(y),
                _={"type": typ},
            )
        )

    def _handle_LangName(self, value: str):
        names = shlex.split(value)
        lang = names.pop(0)
        for nameid, name in enumerate(names):
            if name == "":
                continue
            field = OPENTYPE_NAMES[nameid]
            getattr(self.font.names, field).set_default(name)


class FontForgeSFD(FontForgeSFDIR):
    suffix = ".sfd"

    def _load(self):
        self._setupfont()
        self.current_glyph = None

        self.info = open(Path(self.filename)).read().splitlines()
        self.glyph_order = {}
        self._probe_glyphorder(self.info)
        for gid in sorted(self.glyph_order.keys()):
            self.font.glyphs.append(Glyph(name=self.glyph_order[gid]))
        self._load_line("info")
        while self.info:
            self._load_line("glyph")
        return self.font



# from fontFeatures import Attachment, Routine, Positioning, ValueRecord
import logging
from fontTools.feaLib.builder import addOpenTypeFeaturesFromString
from fontTools.ttLib import newTable
from fontTools.ttLib.tables import otBase, otTables

logger = logging.getLogger(__name__)


def build_all_features(font, ttFont):
    logger.info("Generating opentype features")

    # build_cursive(font)
    # build_mark_mkmk(font)
    # build_mark_mkmk(font, "mkmk")
    # build_kern(font)

    features = font.features.to_fea()
    logger.info("Compiling opentype features")
    addOpenTypeFeaturesFromString(ttFont, features)
    add_gdef_classdef(font, ttFont)


CATMAP = {"base": 1, "ligature": 2, "mark": 3, "component": 4}


def add_gdef_classdef(font, ttFont):
    if not "GDEF" in ttFont:
        ttFont["GDEF"] = newTable("GDEF")
        gdef = otTables.GDEF()
        gdef.Version = 0x00010000
        ttFont["GDEF"].table = gdef
    else:
        gdef = ttFont["GDEF"].table
    if not hasattr(gdef, "GlyphClassDef"):
        gdef.GlyphClassDef = otTables.GlyphClassDef()
        gdef.GlyphClassDef.classDefs = {}
    classdeftable = gdef.GlyphClassDef.classDefs
    if not classdeftable:
        for glyph in font.glyphs:
            if glyph.category in CATMAP:
                classdeftable[glyph.name] = CATMAP[glyph.category]


def build_kern(font):
    def _expand_class(c):
        if c[0] != "@":
            return [c]
        if c[1:] not in font.features.namedClasses:
            logging.info("Attempted to use undefined kerning class %s" % c)
            return None
        return [x for x in font.features.namedClasses[c[1:]] if font.glyphs[x].exported]

    kernroutine = Routine()
    for (left, right), kern in font._all_kerning.items():
        left, right = _expand_class(left), _expand_class(right)
        if left is None or right is None:
            continue
        kernroutine.rules.append(
            Positioning(
                [left, right],
                [ValueRecord(xAdvance=kern), ValueRecord()],
            )
        )
    font.features.addFeature("kern", [kernroutine])


def build_cursive(font):
    anchors = font._all_anchors
    if "entry" in anchors and "exit" in anchors:
        attach = Attachment(
            "entry", "exit", anchors["entry"], anchors["exit"], flags=(0x8 | 0x1)
        )
        r = Routine(
            rules=[attach],
        )
        font.features.addFeature("curs", [r])


def build_mark_mkmk(font, which="mark", strict=False):
    # Find matching pairs of foo/_foo anchors
    anchors = font._all_anchors
    r = Routine(rules=[])
    if which == "mark":
        basecategory = "base"
    else:
        basecategory = "mark"
    for baseanchor in anchors:
        markanchor = "_" + baseanchor
        if markanchor not in anchors:
            continue
        # Filter glyphs to those which are baseanchors
        bases = {
            k: v
            for k, v in anchors[baseanchor].items()
            if font.glyphs[k].exported and (font.glyphs[k].category == basecategory)
        }
        marks = {
            k: v
            for k, v in anchors[markanchor].items()
            if font.glyphs[k].exported
            and (not strict or font.glyphs[k].category == "mark")
        }
        if not (bases and marks):
            continue
        attach = Attachment(baseanchor, markanchor, bases, marks)
        attach.fontfeatures = font.features  # THIS IS A TERRIBLE HACK
        r.rules.append(attach)
    if r.rules:
        font.features.addFeature(which, [r])



import logging

from babelfont import Glyph
from babelfont.Font import Font

logger = logging.getLogger(__name__)


def decompose_mixed_glyphs(font: Font, _args=None):
    logger.info("Decomposing mixed glyphs")

    exportable = set(glyph.name for glyph in font.glyphs if glyph.exported)
    done = set()
    for glyph in font.glyphs:
        decompose_a_glyph(font, glyph, exportable, done)


def decompose_a_glyph(font: Font, glyph: Glyph, exportable, done):
    if glyph.name in done:
        return
    for layer in glyph.layers:
        for c in layer.components:
            decompose_a_glyph(font, font.glyphs[c.ref], exportable, done)
        if (layer.paths and layer.components) or any(
            c.ref not in exportable for c in layer.components
        ):
            layer.decompose()
    done.add(glyph.name)



import logging

from fontTools.feaLib import ast
from fontTools.misc.visitor import Visitor

from babelfont.Font import Font
from babelfont.Glyph import GlyphList

logger = logging.getLogger(__name__)


def rename_glyphs(font: Font, args: dict = {}):
    msg = "Renaming glyphs"
    if "mapping" in args:
        mapping = args["mapping"]
    elif "production" in args:
        msg += " to production names"
        mapping = {
            glyph.name: glyph.production_name
            for glyph in font.glyphs
            if glyph.production_name is not None
        }
    else:
        raise ValueError(
            "rename_glyphs requires either a 'mapping' or 'production' argument"
        )
    logger.info(msg)

    # Step 1: features
    # 1a) classes
    for cls in font.features.classes.keys():
        font.features.classes[cls] = [
            mapping.get(glyph, glyph) for glyph in font.features.classes[cls]
        ]
    # 1b) prefixes
    parsed_features = font.features.as_ast(font)
    for prefix in font.features.prefixes.keys():
        font.features.prefixes[prefix] = _rename_fea(
            parsed_features["prefixes"][prefix],
            mapping,
        ).asFea()
    # 1c) features
    font.features.features = [
        (
            featurename,
            _drop_wrapper(_rename_fea(parsed_ast, mapping)).asFea(),
        )
        for featurename, parsed_ast in parsed_features["features"]
    ]
    for glyph in font.glyphs:
        # Step 2: components
        for layer in glyph.layers:
            for component in layer.components:
                component.ref = mapping.get(component.ref, component.ref)
        # Step 3: glyphs
        glyph.name = mapping.get(glyph.name, glyph.name)
    font.glyphs = GlyphList({glyph.name: glyph for glyph in font.glyphs})


class FeaRenameVisitor(Visitor):
    def __init__(self, mapping):
        self.mapping = mapping


@FeaRenameVisitor.register(ast.GlyphName)
def visit(visitor, gn, *args, **kwargs):
    gn.glyph = visitor.mapping.get(gn.glyph, gn.glyph)
    return False


@FeaRenameVisitor.register(ast.GlyphClass)
def visit(visitor, gcd, *args, **kwargs):
    gcd.glyphs = [visitor.mapping.get(glyph, glyph) for glyph in gcd.glyphs]
    return False


@FeaRenameVisitor.register(ast.LigatureSubstStatement)
def visit(visitor, ligature, *args, **kwargs):
    ligature.replacement = visitor.mapping.get(
        ligature.replacement, ligature.replacement
    )
    return True


@FeaRenameVisitor.register(ast.SinglePosStatement)
def visit(visitor, ccp, *args, **kwargs):
    for glyphclass, _valuerecord in ccp.pos:
        visitor.visit(glyphclass)
    return True


@FeaRenameVisitor.register(ast.MarkClassDefinition)
def visit(visitor, mcd, *args, **kwargs):
    visitor.visitAttr(mcd, "anchor", mcd.anchor, *args, **kwargs)
    visitor.visitAttr(mcd, "glyphs", mcd.glyphs, *args, **kwargs)
    return False


def _rename_fea(parsed: ast.Block, mapping, wrap=None):
    FeaRenameVisitor(mapping).visit(parsed)
    return parsed


def _drop_wrapper(parsed: ast.FeatureFile):
    block = ast.Block()
    block.statements = parsed.statements[0].statements
    return block



import logging

from babelfont.Font import Font
from fontTools.misc.timeTools import timestampSinceEpoch
from fontTools.misc.fixedTools import otRound

logger = logging.getLogger(__name__)


def _default(font, table, key, value, round=otRound):
    if (table, key) not in font.custom_opentype_values:
        font.custom_opentype_values[(table, key)] = round(value)
    return font.custom_opentype_values[(table, key)]


def fill_opentype_values(font: Font, args=None):
    """Prepare a font for final compilation by moving values from
    font attributes to the customOpenTypeValues field."""
    logger.info("Filling in OpenType values")

    def _fallback_metric(*metrics):
        for metric in metrics:
            if callable(metric):
                return metric(font)
            if isinstance(metric, (int, float)):
                return metric
            if (
                metric in font.default_master.metrics
                and font.default_master.metrics[metric] is not None
            ):
                return int(font.default_master.metrics[metric])
        return 0

    version_decimal = font.version[0] + font.version[1] / 10 ** len(
        str(font.version[1])
    )
    _default(font, "head", "fontRevision", version_decimal)
    _default(font, "head", "created", timestampSinceEpoch(font.date.timestamp()))
    _default(font, "head", "lowestRecPPEM", 10)
    ascender = _default(
        font,
        "hhea",
        "ascent",
        _fallback_metric("hheaAscender", "ascender", font.upm * 0.8),
    )
    descender = _default(
        font,
        "hhea",
        "descent",
        _fallback_metric("hheaDescender", "descender", font.upm * -0.2),
    )
    _default(font, "hhea", "lineGap", _fallback_metric("hheaLineGap"))
    _default(font, "OS/2", "usWinAscent", _fallback_metric("winAscent", ascender))
    _default(font, "OS/2", "usWinDescent", _fallback_metric("winDescent", descender))
    # WinDescent should be positive
    if ("OS/2", "usWinDescent") in font.custom_opentype_values:
        font.custom_opentype_values[("OS/2", "usWinDescent")] = abs(
            font.custom_opentype_values[("OS/2", "usWinDescent")]
        )
    _default(font, "OS/2", "sTypoAscender", _fallback_metric("typoAscender", ascender))
    _default(
        font, "OS/2", "sTypoDescender", _fallback_metric("typoDescender", descender)
    )
    _default(
        font,
        "OS/2",
        "sTypoLineGap",
        _fallback_metric("typoLineGap", "lineGap", font.upm * 0.2),
    )
    x_height = _default(
        font, "OS/2", "sxHeight", _fallback_metric("xHeight", font.upm * 0.5)
    )
    _default(font, "OS/2", "sCapHeight", _fallback_metric("capHeight", font.upm * 0.7))
    _default(
        font,
        "OS/2",
        "yStrikeoutPosition",
        _fallback_metric("strikeoutPosition", x_height * 0.6),
    )
    _default(
        font,
        "post",
        "underlinePosition",
        _fallback_metric("underlinePosition", font.upm * -0.075),
    )
    _default(
        font,
        "post",
        "underlineThickness",
        _fallback_metric("underlineThickness", font.upm * 0.05),
    )
    _default(
        font,
        "OS/2",
        "yStrikeoutSize",
        _fallback_metric("underlineThickness", font.upm * 0.05),
    )



import logging

from babelfont.Font import Font

logger = logging.getLogger(__name__)


def zero_background_width(font: Font, args=None):
    logger.info("Zeroing background width")
    for glyph in font.glyphs:
        for layer in glyph.layers:
            if layer.isBackground:
                layer.width = 0


def decompose_backgrounds(font: Font, args=None):
    logger.info("Decomposing backgroundss")
    for glyph in font.glyphs:
        for layer in glyph.layers:
            if layer.isBackground:
                layer.decompose()
